/*
 *      Xpu - display information about CPU utilization in X
 *                                               by T.Komiya
 */

# include <stdio.h>
# include <stdlib.h>
# include <signal.h>
# include <sys/time.h>
# include <string.h>
# include <X11/Xlib.h>
# include <X11/Xutil.h>
# include <X11/keysym.h>
#ifdef SOLARIS
# include <kstat.h>
# include <sys/sysinfo.h>
# include <sys/ddi.h>
# include <unistd.h>
#elif FREEBSD
# include <sys/param.h>
# include <sys/proc.h>
# include <sys/dkstat.h>
# include <sys/buf.h>
# include <sys/uio.h>
# include <sys/namei.h>
# include <sys/malloc.h>
# include <sys/signal.h>
# include <sys/fcntl.h>
# include <sys/ioctl.h>
# include <sys/sysctl.h>
# include <sys/vmmeter.h>

# include <vm/vm_param.h>

# include <ctype.h>
# include <err.h>
# include <errno.h>
# include <kvm.h>
# include <limits.h>
# include <nlist.h>
# include <paths.h>
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <sysexits.h>
# include <time.h>
# include <unistd.h>
# include <devstat.h>

typedef unsigned long ulong;
#define min(n,m)        (n) < (m) ? (n) : (m)

#elif __MACH__
typedef unsigned long ulong;
#define min(n,m)        (n) < (m) ? (n) : (m)
# include <mach/mach_host.h>
# include <mach/mach_port.h>
#endif

#define MAXPROC                 72
#define MAXPROC_NO              999
#define XCPU_SIZE_X             450
#define XCPU_SIZE_X_FOR_3_OPT   300
#define XCPU_UNIT_SIZE_Y        40
#define XCPU_X1_MARGIN          75
#define XCPU_X2_MARGIN          12
#define XCPU_Y_MARGIN           12

int xcpu_size_x      = XCPU_SIZE_X;
int xcpu_unit_size_y = XCPU_UNIT_SIZE_Y;
int interval         = 200000;         /* 0.2 sec */
int tick             = 5;              /* == 1000000/interval */

volatile sig_atomic_t tx1=0, tx2=0; /* M.Yasugi 2010/10/01 */

int column, row, nproc, win_width, win_height, refresh;
int cx;
int **utr, **ktr, **wtr;
ulong *ou, *ok, *ow;

#ifdef LINUX
ulong *nu, *nk, *nw;
unsigned long long prevtotal;
unsigned long long read_procstat(ulong *, ulong *, ulong *);
#define min(n,m)        (n) < (m) ? (n) : (m)
#endif

char pname[MAXPROC][7], hostname[80];
#ifdef SOLARIS
kstat_ctl_t *kc = NULL;
kstat_t *ksp[MAXPROC];
#elif FREEBSD
struct nlist namelist[] = {
#define	X_CPTIME	0
	{ "_cp_time" },
#define X_SUM		1
	{ "_cnt" },
#define	X_BOOTTIME	2
	{ "_boottime" },
#define X_HZ		3
	{ "_hz" },
#define X_STATHZ	4
	{ "_stathz" },
#define X_NCHSTATS	5
	{ "_nchstats" },
#define	X_INTRNAMES	6
	{ "_intrnames" },
#define	X_EINTRNAMES	7
	{ "_eintrnames" },
#define	X_INTRCNT	8
	{ "_intrcnt" },
#define	X_EINTRCNT	9
	{ "_eintrcnt" },
#define	X_KMEMSTATISTICS	10
	{ "_kmemstatistics" },
#define	X_KMEMBUCKETS	11
	{ "_bucket" },
#ifdef notyet
#define	X_DEFICIT	12
	{ "_deficit" },
#define	X_FORKSTAT	13
	{ "_forkstat" },
#define X_REC		14
	{ "_rectime" },
#define X_PGIN		15
	{ "_pgintime" },
#define	X_XSTATS	16
	{ "_xstats" },
#define X_END		17
#else
#define X_END		14
#endif
#if defined(hp300) || defined(luna68k)
#define	X_HPDINIT	(X_END)
	{ "_hp_dinit" },
#endif
#ifdef mips
#define	X_SCSI_DINIT	(X_END)
	{ "_scsi_dinit" },
#endif
#ifdef tahoe
#define	X_VBDINIT	(X_END)
	{ "_vbdinit" },
#define	X_CKEYSTATS	(X_END+1)
	{ "_ckeystats" },
#define	X_DKEYSTATS	(X_END+2)
	{ "_dkeystats" },
#endif
#ifdef vax
#define X_MBDINIT	(X_END)
	{ "_mbdinit" },
#define X_UBDINIT	(X_END+1)
	{ "_ubdinit" },
#endif
	{ "" },
};

struct _disk {
	long time[CPUSTATES];
	long *xfer;
} cur, last;

int *dr_select, dk_ndrive, hz;
kvm_t *kd;

void dkstats(), kread();
#elif __MACH__                  /* FREEBSD */
mach_port_t mach_port, mach_task;
processor_cpu_load_info_t load;
mach_msg_type_number_t count;
#endif

Display *d;
Window w;
GC gc1;
Font f;
XSetWindowAttributes att;
XWMHints *hints;

char *cpu_stat_n();
void disp_stat(), xdisp_stat(), X_Disp_Init(), ctable_init(), refreshwin(),
    refreshwin_exe(), clearwin(), xeveloop(), hand(), xpu_error(),
    realloc_xtr();

int main(int argc, char **argv)
{
    int i, j;
    struct sigaction act;
    struct itimerval tv;
#ifdef SOLARIS
    kstat_t *k;
    cpu_stat_t *cs;
#elif FREEBSD
    char *memf, *nlistf;
    char errbuf[_POSIX2_LINE_MAX];
#endif

#ifdef SOLARIS    
    nproc = (int)sysconf(_SC_NPROCESSORS_ONLN);
#elif LINUX
    nproc = get_number_of_cpu();
#elif FREEBSD
    nproc = 1;
#elif __MACH__
    mach_port = mach_host_self();
    mach_task = mach_task_self();
    count = PROCESSOR_CPU_LOAD_INFO_COUNT;
    host_processor_info(mach_port, PROCESSOR_CPU_LOAD_INFO,
                        (natural_t *)&nproc, (processor_info_array_t *)&load,
                        &count);
#endif
    column = 1;                 /* default */
    row = (nproc + column - 1) / column;
    
    if (1) {
        char dash[] = {1, 1};

        hostname[0] = 'X';  hostname[1] = 'P';  hostname[2] = 'u';
        if (gethostname(hostname+4, 76) == -1)
            hostname[3] = '\0';
        else
            hostname[3] = '@';
        X_Disp_Init(argc, argv);
        ctable_init();
        XSetDashes(d, gc1, 0, dash, 2);
    }

#ifdef SOLARIS
    if ((kc = kstat_open()) == NULL)
        xpu_error("kstat_open() error");

    for (i = j = 0; i < MAXPROC_NO; i++)
        if ((k = kstat_lookup(kc, NULL, -1, cpu_stat_n(i))) != NULL) {
            sprintf(pname[j], "PU %d", i);
            ksp[j++] = k;
        }
    nproc = j;
#elif LINUX
    for (i = 0; i < nproc; i++)
        sprintf(pname[i], "PU %d", i);
#elif FREEBSD
    memf = nlistf = NULL;

    nproc = 1;
    sprintf(pname[0], "PU 0");
    kd = kvm_openfiles(nlistf, memf, NULL, O_RDONLY, errbuf);
    if (kd == 0) 
        errx(1, "kvm_openfiles: %s", errbuf);

    if ((i = kvm_nlist(kd, namelist)) != 0) {
        if (i > 0) {
            warnx("undefined symbols:");
            for (i = 0; i < sizeof(namelist)/sizeof(namelist[0]); i++)
                if (namelist[i].n_type == 0)
                    fprintf(stderr, " %s", namelist[i].n_name);
            (void)fputc('\n', stderr);
        } else
            warnx("kvm_nlist: %s", kvm_geterr(kd));
        exit(1);
    }
    
    if (namelist[X_STATHZ].n_type != 0 && namelist[X_STATHZ].n_value != 0)
        kread(X_STATHZ, &hz, sizeof(hz));
    if (!hz)
        kread(X_HZ, &hz, sizeof(hz));
#elif __MACH__
    for (i = 0; i < nproc; i++)
        sprintf(pname[i], "PU %d", i);
#endif
    utr = ktr = wtr = NULL;
    realloc_xtr(0);

    ou = (ulong *)malloc(nproc * sizeof(ulong));
    ok = (ulong *)malloc(nproc * sizeof(ulong));
    ow = (ulong *)malloc(nproc * sizeof(ulong));
#ifdef LINUX
    nu = (ulong *)malloc(nproc * sizeof(ulong));
    nk = (ulong *)malloc(nproc * sizeof(ulong));
    nw = (ulong *)malloc(nproc * sizeof(ulong));
#endif
    cx = 0;
    clearwin();               /* to display 'PU#' */

#ifdef SOLARIS
    cs = (cpu_stat_t*)malloc(nproc * sizeof(cpu_stat_t));
    for (i = 0; i < nproc; i++)
        kstat_read(kc, ksp[i], &cs[i]);
    for (i = 0; i < nproc; i++) {
        ou[i] = cs[i].cpu_sysinfo.cpu[CPU_USER];
        ok[i] = cs[i].cpu_sysinfo.cpu[CPU_KERNEL];
        ow[i] = cs[i].cpu_sysinfo.cpu[CPU_WAIT];
    }
#elif LINUX
    prevtotal = read_procstat(ou, ok, ow);
#elif FREEBSD
    kread(X_CPTIME, cur.time, sizeof(cur.time));
    dkstats();
    for (i = 0; i < nproc; i++) {
        ou[i] = (ulong)(cur.time[CP_USER] + cur.time[CP_NICE]);
        ok[i] = (ulong)(cur.time[CP_SYS] + cur.time[CP_INTR]);
        ow[i] = 0;
    }
#elif __MACH__
    count = PROCESSOR_CPU_LOAD_INFO_COUNT;
    host_processor_info(mach_port, PROCESSOR_CPU_LOAD_INFO, (natural_t *)&i,
			(processor_info_array_t *)&load, &count);
    for (i = 0; i < nproc; i++) {
        ou[i] = (ulong)(load[i].cpu_ticks[CPU_STATE_USER]
                        + load[i].cpu_ticks[CPU_STATE_NICE]);
        ok[i] = (ulong)load[i].cpu_ticks[CPU_STATE_SYSTEM];
        ow[i] = load[i].cpu_ticks[CPU_STATE_IDLE];
    }
    vm_deallocate(mach_task, (vm_address_t)load, count);
#endif
    act.sa_handler = hand;
    sigemptyset(&(act.sa_mask));
    sigaddset(&(act.sa_mask), SIGALRM);
    act.sa_flags = 0;
    sigaction(SIGALRM, &act, (struct sigaction *)NULL);
    tv.it_value.tv_sec = 0;
    tv.it_value.tv_usec = interval;
    tv.it_interval.tv_sec = 0;
    tv.it_interval.tv_usec = interval;
    setitimer(ITIMER_REAL, &tv, (struct itimerval *)NULL);
    xeveloop();
}

void realloc_xtr(int old_xcpu_size_x)
{
    int i, j;

    if (old_xcpu_size_x >= xcpu_size_x) return;

    utr = (int **)realloc((void *)utr, xcpu_size_x * sizeof(int *));
    ktr = (int **)realloc((void *)ktr, xcpu_size_x * sizeof(int *));
    wtr = (int **)realloc((void *)wtr, xcpu_size_x * sizeof(int *));
    for (i = old_xcpu_size_x; i < xcpu_size_x; i++) {
        utr[i] = (int *)malloc(nproc * sizeof(int));
        ktr[i] = (int *)malloc(nproc * sizeof(int));
        wtr[i] = (int *)malloc(nproc * sizeof(int));
    }
    for (i = old_xcpu_size_x; i < xcpu_size_x; i++)
        for (j = 0; j < nproc; j++)
            utr[i][j] = ktr[i][j] = wtr[i][j] = 0;
}

void hand()
{
    tx1 = (tx1 + 1) & 0xff;     /* M.Yasugi 2010/10/01 */
}

void hand_org()                 /* M.Yasugi 2010/10/01 */
{
    struct sigaction act;
    struct itimerval tv;
    int i;
#ifdef SOLARIS
    cpu_stat_t cs[MAXPROC];
#elif LINUX
    unsigned long long total;
#endif

    if (XEventsQueued(d, QueuedAlready)) {
        XEvent e;
        e.type = Expose;
        XSendEvent(d, w, False, 0, &e); /* ad hoc patch */
    }

#ifdef SOLARIS
    for (i = 0; i < nproc; i++)
        kstat_read(kc, ksp[i], &cs[i]);
    for (i = 0; i < nproc; i++) {
        utr[cx][i] = (cs[i].cpu_sysinfo.cpu[CPU_USER] - ou[i]) * tick;
        ktr[cx][i] = (cs[i].cpu_sysinfo.cpu[CPU_KERNEL] - ok[i]) * tick;
        wtr[cx][i] = (cs[i].cpu_sysinfo.cpu[CPU_WAIT] - ow[i]) * tick;
    }
#elif LINUX
    {
        double k;
        total = read_procstat(nu, nk, nw);
        k = 100.0 / ((double)(total - prevtotal) / (double)nproc);
        for (i = 0; i < nproc; i++) {
            utr[cx][i] = (nu[i] - ou[i]) * k;
            ktr[cx][i] = (nk[i] - ok[i]) * k;
            wtr[cx][i] = (nw[i] - ow[i]) * k;
//            printf("%llu %f\n", total - prevtotal, k);
//            printf("%d %d %d\n", (nu[i] - ou[i]), (nk[i] - ok[i]), (nw[i] - ow[i]));
        }
    }
#elif FREEBSD
    kread(X_CPTIME, cur.time, sizeof(cur.time));
    dkstats();
    for (i = 0; i < nproc; i++) {
        int state;
        double pct, total;

        total = 0;
        for (state = 0; state < CPUSTATES; state++)
            total += cur.time[state];
        if (total)
            pct = 100 / total;
        else
            pct = 0;
        utr[cx][i] = (ulong)((cur.time[CP_USER] + cur.time[CP_NICE]) * pct);
        ktr[cx][i] = (ulong)((cur.time[CP_SYS] + cur.time[CP_INTR]) * pct);
        wtr[cx][i] = ow[i] = 0;
    }
#elif __MACH__
    count = PROCESSOR_CPU_LOAD_INFO_COUNT;
    host_processor_info(mach_port, PROCESSOR_CPU_LOAD_INFO, (natural_t *)&i,
			(processor_info_array_t *)&load, &count);
    for (i = 0; i < nproc; i++) {
        utr[cx][i] = (ulong)(load[i].cpu_ticks[CPU_STATE_USER]
                             + load[i].cpu_ticks[CPU_STATE_NICE] - ou[i]) * 5;
        ktr[cx][i] = (ulong)(load[i].cpu_ticks[CPU_STATE_SYSTEM] - ok[i]) * 5;
        wtr[cx][i] = ow[i] = 0;
    }
#endif

    if (1)
        xdisp_stat(utr, ktr, wtr);
    else
        disp_stat(utr, ktr, wtr);
#ifdef SOLARIS
    for (i = 0; i < nproc; i++) {
        ou[i] = cs[i].cpu_sysinfo.cpu[CPU_USER];
        ok[i] = cs[i].cpu_sysinfo.cpu[CPU_KERNEL];
        ow[i] = cs[i].cpu_sysinfo.cpu[CPU_WAIT];
    }
#elif LINUX
    {
        unsigned long *tmp;
        tmp = ou; ou = nu; nu = tmp;
        tmp = ok; ok = nk; nk = tmp;
        tmp = ow; ow = nw; nw = tmp;
        prevtotal = total;
    }
#elif __MACH__
    for (i = 0; i < nproc; i++) {
        ou[i] = (ulong)(load[i].cpu_ticks[CPU_STATE_USER]
                        + load[i].cpu_ticks[CPU_STATE_NICE]);
        ok[i] = (ulong)load[i].cpu_ticks[CPU_STATE_SYSTEM];
    }
    vm_deallocate(mach_task, (vm_address_t)load, count);
#endif
}

void xeveloop()
{
    char string[10];
    KeySym key;
    XEvent e;
    struct timespec req[1];

    req[0].tv_sec = 0;
    req[0].tv_nsec =  200000*2;

    while (1) {
        while (XPending(d) == 0) /* M.Yasugi 2010/10/01 */
            for (nanosleep(req, 0); tx2 != tx1; tx2 = (tx2 + 1) & 0xff)
                hand_org();
        XNextEvent(d, &e);
        switch (e.type) {
        case ConfigureNotify :
            if (win_width != e.xconfigure.width) {
                int old_xcpu_size_x = xcpu_size_x;

                xcpu_size_x = e.xconfigure.width / column
                    - XCPU_X1_MARGIN - XCPU_X2_MARGIN;
                win_width = e.xconfigure.width;
                cx = cx % xcpu_size_x;
                realloc_xtr(old_xcpu_size_x);
            }
            if (win_height != e.xconfigure.height) {
                xcpu_unit_size_y
                    = (int)(e.xconfigure.height - XCPU_Y_MARGIN) / row
                        - XCPU_Y_MARGIN;
                win_height = e.xconfigure.height;
            }
            refreshwin();
            break;
        case KeyPress :
            if (XLookupString((XKeyEvent *)&e, string, 10, &key, NULL) == 1)
                switch (key) {
                case XK_q :
#ifdef SOLARIS
                    kstat_close(kc);
#endif
                    exit(0);
                case XK_c:
                    clearwin();
                    break;
                case XK_r:
                    refreshwin();
                    break;
                }
            break;
        case Expose:
        case MapNotify:
            refreshwin();
            break;
/*
        case VisibilityNotify:
            break;
*/
        }
    }
}

char *cpu_stat_n(int n)
{
    static char cpu_stat[] = "cpu_stat000";
    char s[4];

    sprintf(s, "%d", n);
    strcpy(cpu_stat + 8, s);
    return cpu_stat;
}

void disp_stat(int u[], int k[], int wt[])
{
    int i;

    for (i = 0; i < nproc; i++)
        printf("%3d %3d %3d  ", u[i], k[i], wt[i]);
    printf("\n");
}

unsigned long MyColor (Display *display, char *color)
{
    Colormap cmap;
    XColor c0, c1;

    cmap = DefaultColormap (display, 0);
    XAllocNamedColor (display, cmap, color, &c1, &c0);
    return (c1.pixel);
}

unsigned long ctable[5];

void ctable_init()
{
    int i;

    ctable[0] = MyColor(d, "black");
    ctable[1] = MyColor(d, "white");
    ctable[2] = MyColor(d, "cyan");
    ctable[3] = MyColor(d, "yellow");
    ctable[4] = MyColor(d, "red");
}

void xdisp_black_rectangle_1pu(int c, int r)
{
    XSetForeground(d, gc1, ctable[0]);
    XFillRectangle(d, w, gc1,
                   (cx % xcpu_size_x) + XCPU_X1_MARGIN
                   + (XCPU_X1_MARGIN + xcpu_size_x + XCPU_X2_MARGIN) * c,
                   (xcpu_unit_size_y + XCPU_Y_MARGIN) * r + XCPU_Y_MARGIN - 2,
                   cx >= (xcpu_size_x - 50) ? xcpu_size_x - cx : 50,
                   xcpu_unit_size_y + 3);
}

void xdisp_cur_line_1pu(int c, int r)
{
    int Xx = (cx % (xcpu_size_x - 1)) + XCPU_X1_MARGIN + 1
        + (XCPU_X1_MARGIN + xcpu_size_x + XCPU_X2_MARGIN) * c,
        Xy = (xcpu_unit_size_y + XCPU_Y_MARGIN) * r + XCPU_Y_MARGIN;

    XSetForeground(d, gc1, ctable[1]);
    XSetLineAttributes(d, gc1, 1, LineOnOffDash, CapButt, JoinMiter);
    XDrawLine(d, w, gc1, Xx, Xy - 1, Xx, Xy + xcpu_unit_size_y);
    XSetLineAttributes(d, gc1, 1, LineSolid, CapButt, JoinMiter);
}

void xdisp_stat_1pu(int i, int c, int r, int x, int u[], int k[], int wt[])
{
    int Xx, Xy;

    Xx = (x % xcpu_size_x) + XCPU_X1_MARGIN
        + (XCPU_X1_MARGIN + xcpu_size_x + XCPU_X2_MARGIN) * c;
    Xy = (xcpu_unit_size_y  + XCPU_Y_MARGIN) * (r + 1);
    if (wt[i]) {
        XSetForeground(d, gc1, ctable[4]);
        XDrawLine(d, w, gc1, Xx , Xy , Xx,
                  Xy -(int)(min(xcpu_unit_size_y,
                                (k[i]+u[i]+wt[i]) / 100.0 * xcpu_unit_size_y)));
    }
    if (u[i]) {
        XSetForeground(d, gc1, ctable[2]);
        XDrawLine(d, w, gc1, Xx, Xy, Xx,
                  Xy - (int)(min(xcpu_unit_size_y,
                                 (u[i] + k[i]) / 100.0 * xcpu_unit_size_y)));
    }
    if (k[i]) {
        XSetForeground(d, gc1, ctable[3]);
        XDrawLine(d, w, gc1, Xx, Xy, Xx,
                  Xy - (int)(min(xcpu_unit_size_y,
                                 k[i] / 100.0 * xcpu_unit_size_y)));
    }
}

void xdisp_stat(int **u, int **k, int **wt)
{
    int i, c, r;

    if (refresh) {
#undef eval
#ifdef eval
        hrtime_t start, end;
        start = gethrtime();
#endif
        refreshwin_exe();
#ifdef eval
        end = gethrtime();
        printf ("%lld nsec\n", end-start);
#endif
    }

    for (i = 0, c = 0; c < column; c++, r = 0)
        for (r = 0; r < row && i < nproc; r++, i++) {
            xdisp_black_rectangle_1pu(c, r);
            xdisp_cur_line_1pu(c, r);
            xdisp_stat_1pu(i, c, r, cx, u[cx], k[cx], wt[cx]);
        }
    XFlush(d);
    cx = ++cx % xcpu_size_x;
}

void refreshwin()
{
    refresh = 1;
}

void refreshwin_exe()
{
    int i, c, r, x;

    //    XSetForeground(d, gc1, MyColor(d, "black"));
    XSetForeground(d, gc1, ctable[0]);
    XFillRectangle(d, w, gc1, 0, 0,
        (xcpu_size_x + XCPU_X1_MARGIN + XCPU_X2_MARGIN) * column,
        (xcpu_unit_size_y+XCPU_Y_MARGIN)*row + XCPU_Y_MARGIN);
    for (i = 0, c = 0; c < column; c++, r = 0)
        for (r = 0; r < row && i < nproc; r++, i++) {
//            XSetForeground(d, gc1, MyColor(d, "white"));
            XSetForeground(d, gc1, ctable[1]);
            XDrawRectangle(d, w, gc1,
                           XCPU_X1_MARGIN - 3
                           + (XCPU_X1_MARGIN + xcpu_size_x + XCPU_X2_MARGIN)*c,
                           (xcpu_unit_size_y  + XCPU_Y_MARGIN) * (r + 1)
                           - xcpu_unit_size_y - 3,
                           xcpu_size_x + 6,
                           xcpu_unit_size_y + 5);
            XDrawString(d, w, gc1,
                        10 + (XCPU_X1_MARGIN + xcpu_size_x + XCPU_X2_MARGIN)*c,
                        (xcpu_unit_size_y  + XCPU_Y_MARGIN) * (r + 1)
                        - xcpu_unit_size_y / 2 + 7,
                        pname[i],
                        strlen(pname[i]));
            for (x = 0; x < cx; x++)
                xdisp_stat_1pu(i, c, r, x, utr[x], ktr[x], wtr[x]);
            xdisp_black_rectangle_1pu(c, r);
            xdisp_cur_line_1pu(c, r);
            x += x >= (xcpu_size_x - 50) ? xcpu_size_x - x : 50;
            for (; x < xcpu_size_x; ++x)
                xdisp_stat_1pu(i, c, r, x, utr[x], ktr[x], wtr[x]);
        }
    XFlush (d);
    refresh = 0;
}

void clearwin()
{
    int i, j, c, r;

    //    XSetForeground(d, gc1, MyColor(d, "black"));
    XSetForeground(d, gc1, ctable[0]);
    XFillRectangle(d, w, gc1, 0, 0,
        (xcpu_size_x + XCPU_X1_MARGIN + XCPU_X2_MARGIN) * column,
        (xcpu_unit_size_y+XCPU_Y_MARGIN)*row + XCPU_Y_MARGIN);

    for (i = 0; i < xcpu_size_x; i++)
        for (j = 0; j < nproc; j++)
            utr[i][j] = ktr[i][j] = wtr[i][j] = 0;
    for (i = 0, c = 0; c < column; c++, r = 0)
        for (r = 0; r < row && i < nproc; r++, i++) {
//            XSetForeground(d, gc1, MyColor(d, "white"));
            XSetForeground(d, gc1, ctable[1]);
            XDrawRectangle(d, w, gc1,
                           XCPU_X1_MARGIN - 3
                           + (XCPU_X1_MARGIN + xcpu_size_x + XCPU_X2_MARGIN)*c,
                           (xcpu_unit_size_y  + XCPU_Y_MARGIN) * (r+1)
                           - xcpu_unit_size_y - 3,
                           xcpu_size_x + 6,
                           xcpu_unit_size_y + 5);
            XDrawString(d, w, gc1,
                        10 + (XCPU_X1_MARGIN + xcpu_size_x + XCPU_X2_MARGIN)*c,
                        (xcpu_unit_size_y  + XCPU_Y_MARGIN) * (r+1)
                        - xcpu_unit_size_y/2 + 7,
                        pname[i],
                        strlen(pname[i]));
            cx = 0;
            xdisp_cur_line_1pu(c, r);
        }
    XFlush (d);
}

void usage()
{
    printf("Xpu - display information about CPU utilization in X\n");
    printf("                                         by T.Komiya\n\n");
    printf("Usage: xpu [-h] [-2] [-3] [-geometry geom]\n");
    printf("Options:\n");
    printf("    -h   print out this message\n");
    printf("    -2   two column layout\n");
    printf("    -3   three column layout\n");
    printf("    -geometry geom\n"
           "         size and position\n");
    printf("Commands:\n");
    printf("    c    clear\n");
    printf("    r    refresh\n");
    printf("    q    quit\n");
    exit(0);
}

void X_Disp_Init(int argc, char **argv)
{
    int flags, i, x = 64, y = 64;
    unsigned int wd, ht;
    char *arg;
    XSizeHints size_hints;
    XClassHint classhint;

    d = XOpenDisplay(NULL);

    for (i = 1, arg = argv[i]; i < argc; i++, arg = argv[i])
        if (arg[0] == '-') {
            switch (arg[1]) {
            case '3' :
            case '2' :
                if (arg[2] == '\0') {
                    if (arg[1] == '3')
                        xcpu_size_x = XCPU_SIZE_X_FOR_3_OPT;
                    column = arg[1] - '0';
                    row = (nproc + column - 1) / column;
                }
                break;
            }
        }
    wd = win_width = (xcpu_size_x + XCPU_X1_MARGIN + XCPU_X2_MARGIN) * column;
    ht = win_height = (xcpu_unit_size_y + XCPU_Y_MARGIN) * row + XCPU_Y_MARGIN;
    for (i = 1, arg = argv[i]; i < argc; i++, arg = argv[i])
        if (arg[0] == '-') {
            switch (arg[1]) {
            case '2':
            case '3':
                if (arg[2] == '\0') break;
            case 'g' :
                if (!strncmp(arg+2, "eometry", 7)) {
                    char s[80];

                    if (i + 1 < argc) {
                        sprintf(s, "=%s", argv[++i]);
                        flags = XParseGeometry(s, &x, &y, &wd, &ht);
                        if (WidthValue & flags) {
                            xcpu_size_x = wd/column - XCPU_X1_MARGIN - XCPU_X2_MARGIN;
                            win_width = wd;
                        }
                        if (HeightValue & flags) {
                            xcpu_unit_size_y = (int)(ht - XCPU_Y_MARGIN) / row
                                                    - XCPU_Y_MARGIN;
                            win_height = ht;
                        }
                        if (XValue & flags)
                            if (XNegative & flags)
                                x = DisplayWidth(d, 0) + x - win_width;
                        if (YValue & flags)
                            if (YNegative & flags)
                                y = DisplayHeight(d, 0) + y - win_height;
                    }
                } else
                    usage();
                break;
            case 'h':
            default :
                usage();
            }
        }

    w = XCreateSimpleWindow(d, RootWindow(d, 0), x, y,
                            win_width, win_height, 2, 0, 1);

    /* set class hint */
    classhint.res_name = "xpu";
    classhint.res_class = "XPu";
    XSetClassHint(d, w, &classhint);

    /* set window group leader to self */
    hints = XAllocWMHints();
    hints->window_group = w;
    hints->flags = WindowGroupHint;
    XSetWMHints(d, w, hints);

    att.backing_store = Always;
    XChangeWindowAttributes(d, w, CWBackingStore, &att);
    XSelectInput(d, w, StructureNotifyMask | KeyPressMask | ExposureMask);

    size_hints.flags = USPosition | USSize;
    size_hints.x = x;       size_hints.y = y;
    size_hints.width = wd;  size_hints.height = ht;
    XSetStandardProperties(d, w, hostname, "xpu", None, argv, argc,
                           &size_hints);

    XSetCommand(d, w, argv, argc);
    XMapWindow(d, w);
    
    gc1 = XCreateGC(d, w, 0, 0);

    //    XSetForeground(d, gc1, MyColor(d, "white"));
    XSetForeground(d, gc1, ctable[1]);
    f = XLoadFont(d, "-*-times-medium-r-normal--17-*");
    XSetFont(d, gc1, f);

/*    refreshwin(); */
}

void xpu_error(char *s)
{
    fprintf(stderr, "xpu: %s\n", s);
#ifdef SOLARIS
    if (kc) kstat_close(kc);
#endif
    exit(1);
}

#ifdef LINUX
int get_number_of_cpu()
{
    FILE *fp;
    char s[80];
    int i;

    if ((fp = fopen("/proc/stat", "r")) == NULL) {
        fprintf(stderr, "Cannot open /proc/stat\n");
        exit(1);
    }
    for (i = 0; fgets(s, 80, fp) != NULL;)
        if (strncmp(s, "cpu ", 4) == 0)
            continue;
        else if (strncmp(s, "cpu", 3) == 0)
            i++;
    fclose(fp);
    return i;
}

unsigned long long read_procstat(ulong *u, ulong *k, ulong *w)
{
    FILE *fp;
    char s[80];
    int i, n;
    unsigned long long usr, nic, sys, idl, iow, tl;

    if ((fp = fopen("/proc/stat", "r")) == NULL) {
        fprintf(stderr, "Cannot open /proc/stat\n");
        exit(1);
    }
    for (i = 0; i < nproc && fgets(s, 80, fp) != NULL;)
        if (strncmp(s, "cpu ", 4) == 0) {
            iow = 0;
            sscanf(s + 5, "%llu %llu %llu %llu %llu", &usr, &nic, &sys, &idl, &iow);
            tl = usr + nic + sys;
            tl += idl + iow;
        } else if (strncmp(s, "cpu", 3) == 0) {
            iow = 0;
            sscanf(s + 3, "%d %llu %llu %llu %llu %llu", &n, &usr, &nic, &sys, &idl, &iow);
            u[n] = usr + nic;
            k[n] = sys;
            w[n] = iow;
            i++;
        }
    fclose(fp);
    return tl;
}
#endif

#ifdef FREEBSD
void
kread(nlx, addr, size)
	int nlx;
	void *addr;
	size_t size;
{
	char *sym;

	if (namelist[nlx].n_type == 0 || namelist[nlx].n_value == 0) {
		sym = namelist[nlx].n_name;
		if (*sym == '_')
			++sym;
		errx(1, "symbol %s not defined", sym);
	}
	if (kvm_read(kd, namelist[nlx].n_value, addr, size) != size) {
		sym = namelist[nlx].n_name;
		if (*sym == '_')
			++sym;
		errx(1, "%s: %s", sym, kvm_geterr(kd));
	}
}

void
dkstats()
{
	register int dn, state;
	double etime;
	long tmp;

	for (dn = 0; dn < dk_ndrive; ++dn) {
		tmp = cur.xfer[dn];
		cur.xfer[dn] -= last.xfer[dn];
		last.xfer[dn] = tmp;
	}
	etime = 0;
	for (state = 0; state < CPUSTATES; ++state) {
		tmp = cur.time[state];
		cur.time[state] -= last.time[state];
		last.time[state] = tmp;
		etime += cur.time[state];
	}
	if (etime == 0)
		etime = 1;
	etime /= hz;
	for (dn = 0; dn < dk_ndrive; ++dn) {
		if (!dr_select[dn])
			continue;
		(void)printf("%2.0f ", cur.xfer[dn] / etime);
	}
}
#endif
